﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Collections;
using Hidori.Windows.Forms;
using System.IO;

namespace SearchEngineListEditorLibrary
{
	public class EditorCore : IObserver
	{
		private TreeView FolderView { get; set; }
		private ListView CurrentItems { get; set; }
		private ToolStripStatusLabel StatusInformation { get; set; }

		private Timer statusMessageRemoveCounter;

		private Options options;

		private ItemTree model;

		public EditorCore( TreeView folderTree, ListView currentItems, ToolStripStatusLabel statusInformation )
		{
			this.FolderView = folderTree;
			this.CurrentItems = currentItems;
			this.StatusInformation = statusInformation;

			Initialize();
		}

		~EditorCore()
		{
			Options.SaveToFile( Options.OptionsFileName, this.options );
		}

		private void Initialize()
		{
			InitializeStatusMessageRemoveCounter();

			InitializeOptions();

			InitializeModel();
		}

		private void InitializeStatusMessageRemoveCounter()
		{
			this.statusMessageRemoveCounter = new Timer();

			this.statusMessageRemoveCounter.Enabled = false;

			this.statusMessageRemoveCounter.Interval = 5000;

			this.statusMessageRemoveCounter.Tick += new EventHandler( statusMessageRemoveCounter_Tick );
		}

		void statusMessageRemoveCounter_Tick( object sender, EventArgs e )
		{
			this.StatusInformation.Text = string.Empty;

			this.statusMessageRemoveCounter.Stop();
		}

		private void LoadMostRecentUsedFile()
		{
			if ( File.Exists( this.options.MostRecentUsedFile ) )
			{
				this.ValidCheckAndDoOpenFile( this.options.MostRecentUsedFile );
			}
			else
			{
				this.options.MostRecentUsedFile = string.Empty;

				this.CreateFolderTree();
			}
		}

		private void InitializeModel()
		{
			this.model = new ItemTree();

			this.model.AddObserver( this );

			LoadMostRecentUsedFile();
		}

		private void InitializeOptions()
		{
			if ( File.Exists( Options.OptionsFileName ) )
			{
				this.options = Options.LoadFromFile( Options.OptionsFileName );
			}
			else
			{
				this.options = new Options();
			}
		}

		private Folder GetFolder( TreeNode node )
		{
			if ( node == null )
			{
				throw new ArgumentNullException();
			}

			if ( node.Tag == null )
			{
				throw new ArgumentException( "node.Tag == null" );
			}

			return ( Folder )node.Tag;
		}

		private Folder GetCurrentFolder()
		{
			if ( this.FolderView.SelectedNode == null )
			{
				return null;
			}
			else
			{
				return GetFolder( this.FolderView.SelectedNode );
			}
		}

		private void NullCheckForCurrentFolderAndDo( Action<Folder> action )
		{
			Folder currentFolder = this.GetCurrentFolder();

			if ( currentFolder != null )
			{
				action( currentFolder );
			}
		}

		public void NewSearchEngine()
		{
			NullCheckForCurrentFolderAndDo( this.DoNewSearchEngine );
		}

		public void NewFolder()
		{
			NullCheckForCurrentFolderAndDo( this.DoNewFolder );
		}

		public void NewGroup()
		{
			NullCheckForCurrentFolderAndDo( this.DoNewGroup );
		}

		public void NewBar()
		{
			NullCheckForCurrentFolderAndDo( this.DoNewBar );
		}

		private void DoNewSearchEngine( Folder currentFolder )
		{
			currentFolder.NewSearchEngine();
		}

		private void DoNewFolder( Folder currentFolder )
		{
			currentFolder.NewFolder();
		}

		private void DoNewGroup( Folder currentFolder )
		{
			currentFolder.NewGroup();
		}

		private void DoNewBar( Folder currentFolder )
		{
			currentFolder.NewBar();
		}

		#region MoveSelectedItems Methods
		private delegate void MoveItems( params int[] indices );

		private void DoMoveItems( MoveItems moveItems )
		{
			if ( this.CurrentItems.SelectedItems.Count <= 0 )
			{
				return;
			}

			this.CurrentItems.BeginUpdate();

			List<Item> selectedItems = GetItemsFromListViewItems( this.CurrentItems.SelectedItems );

			int[] selectedIndices = GetSelectedIndices();

			moveItems( selectedIndices );

			SelectItems( selectedItems );

			this.CurrentItems.EndUpdate();
		}

		public void MoveUpSelectedItems()
		{
			this.DoMoveItems( this.MoveUpItems );
		}

		private void MoveUpItems( params int[] indices )
		{
			this.GetCurrentFolder().MoveUpItems( indices );
		}

		private void SelectItems( List<Item> targetItems )
		{
			foreach ( Item item in targetItems )
			{
				ListViewItem listViewItem = FindListViewItem( item );

				if ( listViewItem != null )
				{
					listViewItem.Selected = true;
				}
			}
		}

		private ListViewItem FindListViewItem( Item item )
		{
			foreach ( ListViewItem listViewItem in this.CurrentItems.Items )
			{
				if ( listViewItem.Tag == item )
				{
					return listViewItem;
				}
			}

			return null;
		}

		private List<Item> GetItemsFromListViewItems( IEnumerable listViewItems )
		{
			List<Item> result = new List<Item>();

			foreach ( ListViewItem item in listViewItems )
			{
				result.Add( ( Item )item.Tag );
			}

			return result;
		}

		public void MoveDownSelectedItems()
		{
			this.DoMoveItems( this.MoveDownItems );
		}

		private void MoveDownItems( params int[] indices )
		{
			this.GetCurrentFolder().MoveDownItems( indices );
		}
		#endregion

		public void Exit()
		{
			Environment.Exit( 0 );
		}

		private void UpdateFolderView( TreeNode topNode )
		{
			DoUpdateFolderTree( topNode );

			UpdateCurrentItems();
		}

		private void DoUpdateFolderTree( TreeNode topNode )
		{
			FolderView.BeginUpdate();

			this.FolderView.Nodes.Clear();

			this.FolderView.Nodes.Add( topNode );

			FolderView.EndUpdate();
		}

		#region CreateFolderTree

		private void CreateFolderTree()
		{
			TreeNode top = DoCreateFolderView();

			UpdateFolderView( top );
		}

		private TreeNode DoCreateFolderView()
		{
			TreeNode top = CreateTreeNode( this.model.Root );

			CreateChildrenAndAddToCurrent( this.model.Root, top );

			return top;
		}

		private void CreateChildrenAndAddToCurrent( Folder folder, TreeNode current )
		{
			foreach ( Item item in folder )
			{
				if ( item is Folder )
				{
					TreeNode node = CreateTreeNode( item );

					current.Nodes.Add( node );

					CreateChildrenAndAddToCurrent( ( Folder )item, node );
				}
			}
		}

		#region CreateTreeNode
		private TreeNode CreateTreeNode( Item item )
		{
			TreeNode node = new TreeNode( item.Name );

			node.Tag = item;

			SetIcons( node );

			return node;
		}

		private void SetIcons( TreeNode node )
		{
			node.ImageIndex = 1;
			node.SelectedImageIndex = 1;
		}
		#endregion

		#endregion

		private void UpdateCurrentItems()
		{
			Folder currentFolder = this.GetCurrentFolder();

			if ( currentFolder == null )
			{
				this.CurrentItems.Items.Clear();
			}
			else
			{
				DoUpdateCurrentItems( currentFolder );
			}
		}

		private void DoUpdateCurrentItems( Folder currentFolder )
		{
			this.CurrentItems.BeginUpdate();

			this.CurrentItems.Items.Clear();

			foreach ( Item item in currentFolder )
			{
				ListViewItem listViewItem = CreateListViewItem( item );

				this.CurrentItems.Items.Add( listViewItem );
			}

			this.CurrentItems.EndUpdate();
		}

		private ListViewItem CreateListViewItem( Item item )
		{
			ListViewItem result = new ListViewItem( item.Name );

			result.Tag = item;

			SetIcons( result, item );

			return result;
		}

		private void SetIcons( ListViewItem listViewItem, Item item )
		{
			if ( item is SearchEngine )
			{
				listViewItem.ImageIndex = 7;
			}
			else if ( item is Folder )
			{
				listViewItem.ImageIndex = 1;
			}
			else if ( item is Group )
			{
				listViewItem.ImageIndex = 8;
			}
			else
			{
				listViewItem.ImageIndex = -1;
			}
		}

		#region CreateDialog Methods
		public static OpenFileDialog CreateOpenFileDialog()
		{
			OpenFileDialog openFileDialog = new OpenFileDialog();

			SetFileDialogWithIniFileMode( openFileDialog );

			return openFileDialog;
		}

		public static SaveFileDialog CreateSaveFileDialog()
		{
			SaveFileDialog dialog = new SaveFileDialog();

			SetFileDialogWithIniFileMode( dialog );

			return dialog;
		}

		private const string iniFileExtensionFilter = @"iniファイル (*.ini)|*.ini";

		private static void SetFileDialogWithIniFileMode( FileDialog dialog )
		{
			dialog.Filter = iniFileExtensionFilter;

			dialog.DefaultExt = @"ini";
		}
		#endregion

		private void ShowStatusMessage( string message )
		{
			this.StatusInformation.Text = message;

			ReStartStatusMessageRemoveCounter();
		}

		private void ReStartStatusMessageRemoveCounter()
		{
			this.statusMessageRemoveCounter.Stop();

			this.statusMessageRemoveCounter.Start();
		}

		#region OpenFile
		public void OpenFile()
		{
			OpenFileDialog openFileDialog = CreateOpenFileDialog();

			if ( openFileDialog.ShowDialog() == DialogResult.OK )
			{
				this.ValidCheckAndDoOpenFile( openFileDialog.FileName );
			}
		}

		public void ValidCheckAndDoOpenFile( string fileName )
		{
			if ( SearchEngineIniFile.IsValidSearchIniFile( fileName ) )
			{
				DoOpenFile( fileName );
			}
			else
			{
				System.Windows.Forms.MessageBox.Show( fileName + "は無効なファイルです。" );
			}
		}

		public void DoOpenFile( string fileName )
		{
			this.model.LoadFromFile( fileName );

			this.options.MostRecentUsedFile = fileName;

			this.FolderView.ExpandAll();

			this.FolderView.SelectedNode = this.FolderView.TopNode;

			ShowStatusMessage( fileName + "を開きました" );
		}

		#endregion

		#region SaveFile

		public void OverWriteSaveFile()
		{
			if ( File.Exists( this.options.MostRecentUsedFile ) )
			{
				DoSaveFile( this.options.MostRecentUsedFile );
			}
			else
			{
				this.SaveFile();
			}
		}

		public void SaveFile()
		{
			SaveFileDialog saveFileDialog = CreateSaveFileDialog();

			if ( saveFileDialog.ShowDialog() == DialogResult.OK )
			{
				DoSaveFile( saveFileDialog.FileName );
			}
		}

		public void DoSaveFile( string fileName )
		{
			this.model.SaveToFile( fileName );

			this.options.MostRecentUsedFile = fileName;

			ShowStatusMessage( fileName + "に保存しました" );
		}

		#endregion

		public List<Item> GetSelectedItems()
		{
			List<Item> result = new List<Item>();

			foreach ( ListViewItem listViewItem in this.CurrentItems.SelectedItems )
			{
				result.Add( ( Item )listViewItem.Tag );
			}

			return result;
		}

		private List<Item> copyTargets;

		public void CopyItems()
		{
			copyTargets = this.GetSelectedItems();

			NullCheckForCurrentFolderAndDo( this.DoCopyItems );
		}

		private void DoCopyItems( Folder currentFolder )
		{
			foreach ( Item item in copyTargets )
			{
				Item clone = ( Item )item.Clone();

				currentFolder.Add( clone );
			}
		}

		#region EditItem
		public void EditItem()
		{
			if ( HasSelectedOnlyOneItem() )
			{
				Item selectedItem = GetSelectedItem();

				if ( IsEditableItem( selectedItem ) )
				{
					DoEditItem( selectedItem );
				}
			}
		}

		private bool HasSelectedOnlyOneItem()
		{
			return this.CurrentItems.SelectedItems.Count == 1;
		}

		private Item GetSelectedItem()
		{
			if ( this.CurrentItems.SelectedItems.Count == 1 )
			{
				return ( Item )this.CurrentItems.SelectedItems[0].Tag;
			}

			return null;
		}

		private bool IsEditableItem( Item selectedItem )
		{
			return !( selectedItem is Bar );
		}

		public void DoEditItem( Item item )
		{
			using ( EditForm editForm = CreateEditForm( item ) )
			{
				editForm.ShowDialog();
			}
		}

		public static EditForm CreateEditForm( Item item )
		{
			EditForm editForm = new EditForm( item );

			if ( item is SearchEngine )
			{
				editForm.SetSearchEngineEditMode();
			}
			else if ( item is Folder )
			{
				editForm.SetFolderEditMode();
			}
			else if ( item is Group )
			{
				editForm.SetGroupEditMode();
			}
			else
			{
				throw new ArgumentException();
			}

			return editForm;
		}

		public bool HasSelectedItem()
		{
			return 0 < CurrentItems.SelectedItems.Count;
		}
		#endregion

		#region RemoveSelectedItems

		private int[] removeTargetsIndices;

		public void RemoveSelectedItems()
		{
			removeTargetsIndices = GetSelectedIndices();

			NullCheckForCurrentFolderAndDo( this.DoRemoveSelectedItmes );
		}

		private void DoRemoveSelectedItmes( Folder currentFolder )
		{
			for ( int index = removeTargetsIndices.Length - 1; 0 <= index; index-- )
			{
				currentFolder.RemoveAt( removeTargetsIndices[index] );
			}
		}

		private int[] GetSelectedIndices()
		{
			int[] result = new int[this.CurrentItems.SelectedIndices.Count];

			this.CurrentItems.SelectedIndices.CopyTo( result, 0 );

			return result;
		}
		#endregion

		public void FolderTreeAfterSelect()
		{
			this.UpdateCurrentItems();
		}

		private TreeNode GetTreeNode( Folder folder )
		{
			foreach ( TreeNode node in this.FolderView.Nodes )
			{
				if ( node.Tag == folder )
				{
					return node;
				}
			}

			return null;
		}

		#region IObserver メンバ

		// Itemごとに処理を分けようかと思っていたけど面倒だから毎回全部作り直し
		public void Update( Item item )
		{
			ViewState backup = this.CreateBackupState( this );

			this.CreateFolderTree();

			this.RestoreBackupState( backup );
		}

		#endregion

		#region Backup and Restore
		private ViewState CreateBackupState( EditorCore editorCore )
		{
			ViewState state = new ViewState();

			BackupSelectedNodeState( state );

			BackupCollapsedNodesState( state );

			BackupFocusedListViewItemState( state );

			BackupSelectedListViewItemsState( state );

			return state;
		}

		private void BackupSelectedListViewItemsState( ViewState state )
		{
			state.SelectedListViewItemsAttachedItems = GetItemsFromListViewItems( this.CurrentItems.SelectedItems );
		}

		private void BackupFocusedListViewItemState( ViewState state )
		{
			if ( this.CurrentItems.FocusedItem != null )
			{
				state.FocusedListViewItemAttachedItem = ( Item )this.CurrentItems.FocusedItem.Tag;
			}
		}

		private bool IsCollapsed( TreeNode node )
		{
			return !node.IsExpanded;
		}

		private void BackupCollapsedNodesState( ViewState state )
		{
			TreeNode[] collapsedNodes =
				this.FolderView.Nodes.FindAllRecursive( this.IsCollapsed );

			List<Item> items = new List<Item>();

			foreach ( TreeNode node in collapsedNodes )
			{
				items.Add( ( Item )node.Tag );
			}

			state.CollapsedNodesAttachedItems = items;
		}

		private void BackupSelectedNodeState( ViewState state )
		{
			if ( this.HasSelectedNode() )
			{
				state.SelectedNodeAttachedItem = ( Item )this.FolderView.SelectedNode.Tag;
			}
		}

		private bool HasSelectedNode()
		{
			return this.FolderView.SelectedNode != null;
		}

		private void RestoreBackupState( ViewState state )
		{
			this.FolderView.ExpandAll();

			RestoreCollapsedNodes( state );

			RestoreSelectedNode( state );

			RestoreSelectedListViewItems( state );

			RestoreFocusedListViewItem( state );
		}

		private void RestoreSelectedListViewItems( ViewState state )
		{
			foreach ( Item item in state.SelectedListViewItemsAttachedItems )
			{
				ListViewItem selectTarget = this.FindListViewItem( item );

				if ( selectTarget != null )
				{
					selectTarget.Selected = true;
				}
			}
		}

		private void RestoreFocusedListViewItem( ViewState state )
		{
			if ( state.FocusedListViewItemAttachedItem != null )
			{
				ListViewItem focusTarget = this.FindListViewItem( state.FocusedListViewItemAttachedItem );

				if ( focusTarget != null )
				{
					focusTarget.Focused = true;
				}
			}
		}

		private void RestoreCollapsedNodes( ViewState state )
		{
			foreach ( Item item in state.CollapsedNodesAttachedItems )
			{
				TreeNode node = FindNodeFromItem( item );

				if ( node != null )
				{
					node.Collapse();
				}
			}
		}

		private void RestoreSelectedNode( ViewState state )
		{
			if ( state.SelectedNodeAttachedItem != null )
			{
				TreeNode node = FindNodeFromItem( state.SelectedNodeAttachedItem );

				if ( node != null )
				{
					this.FolderView.SelectedNode = node;
				}
			}
		}

		private Item findTargetItem;

		private bool HasMatchItem( TreeNode node )
		{
			return node.Tag == findTargetItem;
		}

		private TreeNode FindNodeFromItem( Item item )
		{
			findTargetItem = item;

			return this.FolderView.Nodes.FindRecursive( this.HasMatchItem );
		}
		#endregion

		#region ShowShortCutList
		public void ShowShortCutList()
		{
			List<Item> list = Folder.FindAll( this.model.Root, this.HasShortCut, true );

			StringBuilder builder = new StringBuilder();

			for ( int i = 1; i <= 26; i++ )
			{
				specialShortCut = i;
				List<Item> temp = list.FindAll( this.HasSpecialShortCut );

				switch ( temp.Count )
				{
					case 0:
						AppendNoAssignedShortCutMessage( builder, i );
						break;

					case 1:
						AppendValidAssignedShortCutMessage( builder, i, temp[0] );
						break;

					default:
						AppendOverLappedAssignedShortCutMessages( builder, i, temp );
						break;
				}
			}

			MessageBox.Show( builder.ToString() );
		}

		private void AppendNoAssignedShortCutMessage( StringBuilder builder, int shortCut )
		{
			builder.Append( CreateMessage( shortCut, null, false ) );
		}

		private void AppendValidAssignedShortCutMessage( StringBuilder builder, int shortCut, Item item )
		{
			builder.Append( CreateMessage( shortCut, item, false ) );
		}

		private void AppendOverLappedAssignedShortCutMessages( StringBuilder builder, int shortCut, List<Item> overLappedItems )
		{
			foreach ( Item item in overLappedItems )
			{
				builder.Append( CreateMessage( shortCut, item, true ) );
			}
		}

		private int specialShortCut;

		private bool HasSpecialShortCut( Item item )
		{
			return item.ShortCut == specialShortCut;
		}

		private string CreateMessage( int shortCut, Item item, bool overLapped )
		{
			if ( overLapped )
			{
				return DoCreateMessage( shortCut, "重複    " + item.Name );
			}
			else
			{
				if ( item == null )
				{
					return DoCreateMessage( shortCut, "未割り当て" );
				}
				else
				{
					return DoCreateMessage( shortCut, item.Name );
				}
			}
		}

		private string DoCreateMessage( int shortCut, string message )
		{
			return EditorCore.ShortCutNumberToAlphabet( shortCut ) +
				" - " + message + "\n";
		}

		private bool HasShortCut( Item item )
		{
			return 0 < item.ShortCut;
		}
		#endregion

		#region ConvertAlphabetToShortCutNumber Methods
		public static string ShortCutNumberToAlphabet( int shortCut )
		{
			if ( shortCut < 1 || 26 < shortCut )
			{
				throw new ArgumentOutOfRangeException();
			}

			string result = Convert.ToChar( shortCut + codeBasicPoint ).ToString();

			return result;
		}

		static int codeBasicPoint = 96;

		public static int AlphabetToShortCutNumber( string alphabet )
		{
			if ( alphabet.Length != 1 )
			{
				throw new ArgumentException();
			}

			int shortCut = Convert.ToInt32( Convert.ToChar( alphabet ) ) - codeBasicPoint;

			if ( shortCut < 1 || 26 < shortCut )
			{
				throw new ArgumentException();
			}

			return shortCut;
		}
		#endregion
	}
}
