﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Tomham.IO;
using Tomham.Collections.Generic;

namespace SearchEngineListEditorLibrary
{
	public class Folder : Item, IEnumerable<Item>, IList<Item>
	{
		public Folder()
			: base()
		{
			this.children = new List<Item>();
		}

		private List<Item> children { get; set; }

		public const string RootName = "Root";

		internal override void SetOwner( ItemTree owner )
		{
			base.SetOwner( owner );

			foreach ( Item item in this.children )
			{
				item.SetOwner( owner );
			}
		}

		public override void CreateAndAddRawItems( List<Item> rawItems )
		{
			rawItems.Add( CreateFolderFirstRawItem() );

			foreach ( Item item in this )
			{
				item.CreateAndAddRawItems( rawItems );
			}

			rawItems.Add( CreateFolderLastRawItem() );
		}

		private Item CreateFolderFirstRawItem()
		{
			return CreateRawItem( "[" + this.Name );
		}

		private Item CreateFolderLastRawItem()
		{
			return CreateRawItem( "]" );
		}

		private Item CreateRawItem( string name )
		{
			return new Item { Name = name };
		}

		public void NewSearchEngine()
		{
			this.Add( new SearchEngine { Name = "NewSearchEngine" } );
		}

		public void NewFolder()
		{
			this.Add( new Folder { Name = "NewFolder" } );
		}

		public void NewGroup()
		{
			Group group = CreateGroup();

			this.Add( group );
		}

		private Group CreateGroup()
		{
			Group group = new Group { Name = "NewGroup" };

			List<SearchEngine> searchEngines = GetSearchEngines();

			SetAllSearchEngineNames( group, searchEngines );

			return group;
		}

		private void SetAllSearchEngineNames( Group group, List<SearchEngine> searchEngines )
		{
			List<string> searchEngineNames = new List<string>();

			foreach ( SearchEngine searchEngine in searchEngines )
			{
				searchEngineNames.Add( searchEngine.Name );
			}

			group.SearchEngineNames = searchEngineNames;
		}

		private List<SearchEngine> GetSearchEngines()
		{
			List<SearchEngine> result = new List<SearchEngine>();

			foreach ( Item item in this )
			{
				if ( item is SearchEngine )
				{
					result.Add( ( SearchEngine )item );
				}
			}

			return result;
		}

		public void NewBar()
		{
			this.Add( new Bar() );
		}

		#region IEnumerable<Item> メンバ

		public IEnumerator<Item> GetEnumerator()
		{
			return this.children.GetEnumerator();
		}

		#endregion

		#region IEnumerable メンバ

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return this.GetEnumerator();
		}

		#endregion

		#region IList<Item> メンバ

		public int IndexOf( Item item )
		{
			return this.children.IndexOf( item );
		}

		public void Insert( int index, Item item )
		{
			this.children.Insert( index, item );

			item.SetParent( this );

			SetOwnerAndNotifyToOwner( item, this.Owner );
		}

		private void SetOwnerAndNotifyToOwner( Item item, ItemTree owner )
		{
			item.SetOwner( owner );

			NotifyToOwner();
		}

		public void RemoveAt( int index )
		{
			Item item = this.children[index];

			this.Remove( item );
		}

		public Item this[int index]
		{
			get
			{
				return this.children[index];
			}
			set
			{
				this.RemoveAt( index );

				if ( GetLastIndex() < index )
				{
					this.Add( value );
				}
				else
				{
					this.Insert( index, value );
				}
			}
		}

		private int GetLastIndex()
		{
			return this.Count - 1;
		}

		#endregion

		#region ICollection<Item> メンバ

		public void Add( Item item )
		{
			this.children.Add( item );

			item.SetParent( this );

			SetOwnerAndNotifyToOwner( item, this.Owner );
		}

		public void Clear()
		{
			ClearOwnerAndParentForList( this.children );

			this.children.Clear();

			NotifyToOwner();
		}

		private void ClearOwnerAndParentForList( List<Item> list )
		{
			foreach ( Item item in list )
			{
				item.SetOwner( null );

				item.SetParent( null );
			}
		}

		public bool Contains( Item item )
		{
			return this.children.Contains( item );
		}

		public void CopyTo( Item[] array, int arrayIndex )
		{
			this.children.CopyTo( array, arrayIndex );
		}

		public int Count
		{
			get { return this.children.Count; }
		}

		public bool IsReadOnly
		{
			get { return false; }
		}

		public bool Remove( Item item )
		{
			bool result = this.children.Remove( item );

			item.SetParent( null );

			SetOwnerAndNotifyToOwner( item, null );

			return result;
		}

		#endregion

		public Item Find( Predicate<Item> match )
		{
			return this.children.Find( match );
		}

		public void ForEach( Action<Item> action )
		{
			this.children.ForEach( action );
		}

		public static List<Item> FindAll( Folder folder, Predicate<Item> match, bool recursive )
		{
			if ( folder == null ) throw new ArgumentNullException( "folder" );
			if ( match == null ) throw new ArgumentNullException( "match" );

			List<Item> list = new List<Item>();

			foreach ( Item item in folder )
			{
				if ( match( item ) )
				{
					list.Add( item );
				}
			}

			if ( recursive )
			{
				foreach ( Item item in folder )
				{
					if ( item is Folder )
					{
						list.AddRange( FindAll( ( Folder )item, match, recursive ) );
					}
				}
			}

			return list;
		}

		public void MoveUpItems( params int[] indices )
		{
			IListUtility.MoveUpItems( this.children, indices );

			NotifyToOwner();
		}

		public void MoveDownItems( params int[] indices )
		{
			IListUtility.MoveDownItems( this.children, indices );

			NotifyToOwner();
		}

		public List<Item> GetItemList()
		{
			return this.children.GetRange( 0, this.children.Count - 1 );
		}

		public override object Clone()
		{
			Folder result = new Folder();

			result.SetProperties(
				this.Name,
				this.URLFirst,
				this.URLLast,
				this.Lock,
				this.Encode,
				this.ShortCut,
				this.Icon 
				);

			foreach ( Item child in this.children )
			{
				Item clonedChild = ( Item )child.Clone();

				result.Add( clonedChild );
			}

			return result;
		}
	}
}
