﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SearchEngineListEditorLibrary
{
	public class IniFileReader
	{
		public Folder LoadFromFile( string fileName )
		{
			List<Item> rawItems = ReadRawItems( fileName );

			return CreateItemTree( rawItems );
		}

		private List<Item> ReadRawItems( string fileName )
		{
			SearchEngineIniFile iniFile = new SearchEngineIniFile();

			iniFile.LoadFromFile( fileName );

			int itemCount = iniFile.ReadItemCount();

			List<Item> rawItems = new List<Item>();

			for ( int itemNumber = 0; itemNumber < itemCount; itemNumber++ )
			{
				Item item = ReadRawItem( iniFile, itemNumber );

				rawItems.Add( item );
			}

			return rawItems;
		}

		private Item ReadRawItem( SearchEngineIniFile iniFile, int itemNumber )
		{
			Item item = new Item
			{
				Name = iniFile.ReadName( itemNumber ),
				URLFirst = iniFile.ReadURLFirst( itemNumber ),
				URLLast = iniFile.ReadURLLast( itemNumber ),
				Icon = iniFile.ReadIcon( itemNumber ),
				Lock = iniFile.ReadLock( itemNumber ),
				ShortCut = iniFile.ReadShotcut( itemNumber ),
				Encode = iniFile.ReadEncode( itemNumber )
			};

			return item;
		}

		private Folder CreateItemTree( List<Item> rawItems )
		{
			Folder root = new Folder { Name = Folder.RootName };

			Stack<Folder> folders = new Stack<Folder>();

			SetCurrentFolder( root, folders );

			foreach ( Item rawItem in rawItems )
			{
				CreateItemAndAddToTree( rawItem, folders );
			}

			return root;
		}

		private void CreateItemAndAddToTree( Item rawItem, Stack<Folder> folders )
		{
			if ( IsFolderLast( rawItem ) )
			{
				BackToPreviousCurrentFolder( folders );
			}
			else
			{
				DoCreateItemAndAddToTree( rawItem, folders );
			}
		}

		private void DoCreateItemAndAddToTree( Item rawItem, Stack<Folder> folders )
		{
			Item newItem = CreateItem( rawItem );

			AddItemToCurrentFolder( newItem, folders );

			if ( newItem is Folder )
			{
				Folder newCurrentFolder = ( Folder )newItem;
				SetCurrentFolder( newCurrentFolder, folders );
			}
		}

		private void AddItemToCurrentFolder( Item item, Stack<Folder> folders )
		{
			folders.Peek().Add( item );
		}

		private void SetCurrentFolder( Folder newCurrentFolder, Stack<Folder> folders )
		{
			folders.Push( newCurrentFolder );
		}

		private void BackToPreviousCurrentFolder( Stack<Folder> folders )
		{
			folders.Pop();
		}

		private Item CreateItem( Item rawItem )
		{
			Item newItem;

			if ( IsFolderFirst( rawItem ) )
			{
				newItem = CreateFolder( rawItem );
			}
			else if ( IsGroup( rawItem ) )
			{
				newItem = CreateGroup( rawItem );
			}
			else if ( IsBar( rawItem ) )
			{
				newItem = new Bar();
			}
			else
			{
				newItem = CreateSearchEngine( rawItem );
			}

			return newItem;
		}

		private Item CreateSearchEngine( Item rawItem )
		{
			Item item = new SearchEngine();

			rawItem.CopyTo( item );

			return item;
		}

		private Item CreateGroup( Item rawItem )
		{
			Group group = new Group
			{
				Name = rawItem.Name,
				ShortCut = rawItem.ShortCut,
				Icon = rawItem.Icon
			};

			group.SearchEngineNames = GetItemNames( rawItem );

			return group;
		}

		private readonly char pipe = Convert.ToChar( "|" );

		private List<string> GetItemNames( Item rawItem )
		{
			string names = rawItem.URLFirst.Substring( 6 );

			return new List<string>( names.Split( pipe ) );
		}

		private Item CreateFolder( Item rawItem )
		{
			string name = rawItem.Name.Substring( 1 );

			return new Folder { Name = name };
		}

		#region RawItem is SomeItemType Methods

		private bool IsBar( Item rawItem )
		{
			return rawItem.Name == Bar.BarString;
		}

		private bool IsFolderFirst( Item rawItem )
		{
			return rawItem.Name.StartsWith( "[" );
		}

		private bool IsFolderLast( Item rawItem )
		{
			return rawItem.Name.StartsWith( "]" );
		}

		private bool IsGroup( Item rawItem )
		{
			return rawItem.URLFirst.StartsWith( "group:" );
		}

		#endregion
	}
}
