﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SearchEngineListEditorLibrary
{
	public class ItemTree
	{
		private Folder root;

		public Folder Root
		{
			get { return root; }
		}

		public ItemTree()
		{
			SetRoot( new Folder { Name = Folder.RootName } );
		}

		private void SetRoot( Folder root )
		{
			this.root = root;

			SetOwnerRecursive( root );
		}

		private void SetOwnerRecursive( Folder folder )
		{
			folder.SetOwner( this );

			foreach ( Item item in folder )
			{
				if ( item is Folder )
				{
					SetOwnerRecursive( ( Folder )item );
				}
				else
				{
					item.SetOwner( this );
				}
			}
		}

		private List<IObserver> observers = new List<IObserver>();

		public void AddObserver( IObserver observer )
		{
			observers.Add( observer );
		}

		public void DeleteObserver( IObserver observer )
		{
			observers.Remove( observer );
		}

		#region Load and Save File

		public void LoadFromFile( string fileName )
		{
			IniFileReader reader = new IniFileReader();

			Folder newRoot = reader.LoadFromFile( fileName );

			this.SetRoot( newRoot );

			this.NotifyObservers( newRoot );
		}

		public void SaveToFile( string fileName )
		{
			IniFileWriter writer = new IniFileWriter();

			writer.SaveToFile( fileName, this.Root );
		}

		#endregion

		#region Update Methods

		private int updateWaitCount = 0;

		public void BeginUpdate()
		{
			updateWaitCount++;
		}

		public void EndUpdate()
		{
			updateWaitCount--;

			if ( updateWaitCount < 0 )
			{
				string message = "BeginUpdate() called less than EndUpdate() called.";

				throw new InvalidOperationException( message );
			}

			this.Update( this.root );
		}

		internal void Update( Item item )
		{
			if ( CanNotifyObservers() )
			{
				this.NotifyObservers( item );
			}
		}

		private void NotifyObservers( Item item )
		{
			foreach ( IObserver observer in observers )
			{
				observer.Update( item );
			}
		}

		private bool CanNotifyObservers()
		{
			return this.updateWaitCount == 0;
		}

		#endregion

		public static void MoveItem( Item item, Folder destination )
		{
			if ( item == null )
			{
				throw new ArgumentNullException( "item" );
			}

			if ( destination == null )
			{
				throw new ArgumentNullException( "destination" );
			}

			if ( item == destination )
			{
				throw new ArgumentException( "item == destination" );
			}

			if ( item.Parent == destination )
			{
				throw new ArgumentException( "item.Parent == destination" );
			}

			if ( item is Folder && destination.IsAncester( ( Folder )item ) )
			{
				throw new ArgumentException( "item is an ancester of destination." );
			}

			DoMoveItem( item, destination );
		}

		private static void DoMoveItem( Item item, Folder destination )
		{
			if ( item.Parent != null )
			{
				item.Parent.Remove( item );
			}

			destination.Add( item );
		}
	}
}
