﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

using Gajatko.IniFiles.Light;

namespace SearchEngineListEditorLibrary
{
	public class SearchEngineIniFile
	{
		private static string sectionName = "SearchEngine";

		private static string searchEngineCountKey = "SearchEngineCount";

		public static bool IsValidSearchIniFile( string fileName )
		{
			IniFileLight iniFile = new IniFileLight( fileName );

			return iniFile.Sections.ContainsKey( sectionName );
		}

		private IniFileLight iniFile;

		public SearchEngineIniFile()
		{
			Initialize();
		}

		private void Initialize()
		{
			iniFile = new IniFileLight();

			iniFile.Sections.Add( sectionName, new Dictionary<string, string>() );
		}

		#region Save and Load File
		public void LoadFromFile( string fileName )
		{
			if ( File.Exists( fileName ) )
			{
				StreamReader reader = new StreamReader( fileName, Encoding.Default );

				try
				{
					iniFile = new IniFileLight( reader );
				}
				finally
				{
					reader.Close();
				}
			}
			else
			{
				throw new FileNotFoundException();
			}
		}

		public void SaveToFile( string fileName )
		{
			StreamWriter writer = new StreamWriter( fileName, false, Encoding.Default, 100 );

			try
			{
				iniFile.Save( writer );
			}
			finally
			{
				writer.Close();
			}
		} 
		#endregion

		private string ReadString( string key )
		{
			if ( iniFile.Sections[sectionName].ContainsKey( key ) )
			{
				return RemoveQuotaitions( iniFile.Sections[sectionName][key] );
			}
			else
			{
				return string.Empty;
			}
		}

		private char doubleQuotation = Convert.ToChar( "\"" );

		private string RemoveQuotaitions( string value )
		{
			string result = value.Trim( new char[] { doubleQuotation } );

			return result;
		}

		private int ReadInteger( string key )
		{
			string temp = ReadString( key );

			return Convert.ToInt32( temp );
		}

		private bool ReadBoolean( string key )
		{
			string temp = ReadString( key );

			return Convert.ToBoolean( Convert.ToInt32( temp ) );
		}

		private void WriteString( string key, string value )
		{
			iniFile.Sections[sectionName].Add( key, value );
		}

		private void WriteInteger( string key, int value )
		{
			WriteString( key, value.ToString() );
		}

		private void WriteBoolean( string key, bool value )
		{
			int data = Convert.ToInt32( value );

			WriteString( key, data.ToString() );
		}

		#region Read Methods

		public int ReadItemCount()
		{
			return ReadInteger( searchEngineCountKey );
		}

		public string ReadName( int itemNumber )
		{
			return ReadString( CreateNameKey( itemNumber ) );
		}

		public string ReadURLFirst( int itemNumber )
		{
			return ReadString( CreateURLFirstKey( itemNumber ) );
		}

		public string ReadURLLast( int itemNumber )
		{
			return ReadString( CreateURLLastKey( itemNumber ) );
		}

		public int ReadEncode( int itemNumber )
		{
			return ReadInteger( CreateEncodeKey( itemNumber ) );
		}

		public int ReadShotcut( int itemNumber )
		{
			return ReadInteger( CreateShortcutKey( itemNumber ) );
		}

		public bool ReadLock( int itemNumber )
		{
			return ReadBoolean( CreateLockKey( itemNumber ) );
		}

		public int ReadIcon( int itemNumber )
		{
			return ReadInteger( CreateIconKey( itemNumber ) );
		}
		#endregion

		#region Write Methods

		public void WriteItemCount( int count )
		{
			WriteInteger( searchEngineCountKey, count );
		}

		public void WriteName( int itemNumber, string value )
		{
			WriteString( CreateNameKey( itemNumber ), value );
		}

		public void WriteURLFirst( int itemNumber, string value )
		{
			WriteString( CreateURLFirstKey( itemNumber ), value );
		}

		public void WriteURLLast( int itemNumber, string value )
		{
			WriteString( CreateURLLastKey( itemNumber ), value );
		}

		public void WriteEncode( int itemNumber, int value )
		{
			WriteInteger( CreateEncodeKey( itemNumber ), value );
		}

		public void WriteShotcut( int itemNumber, int value )
		{
			WriteInteger( CreateShortcutKey( itemNumber ), value );
		}

		public void WriteLock( int itemNumber, bool value )
		{
			WriteBoolean( CreateLockKey( itemNumber ), value );
		}

		public void WriteIcon( int itemNumber, int value )
		{
			WriteInteger( CreateIconKey( itemNumber ), value );
		}
		#endregion

		#region CreateKeys

		private string CreateNameKey( int itemNumber )
		{
			return CreateKey( itemNumber, "_Name" );
		}

		private string CreateLockKey( int itemNumber )
		{
			return CreateKey( itemNumber, "_Lock" );
		}

		private string CreateURLFirstKey( int itemNumber )
		{
			return CreateKey( itemNumber, "_First" );
		}

		private string CreateURLLastKey( int itemNumber )
		{
			return CreateKey( itemNumber, "_Last" );
		}

		private string CreateShortcutKey( int itemNumber )
		{
			return CreateKey( itemNumber, "_Shortcut" );
		}

		private string CreateEncodeKey( int itemNumber )
		{
			return CreateKey( itemNumber, "_Encode" );
		}

		private string CreateIconKey( int itemNumber )
		{
			return CreateKey( itemNumber, "_Icon" );
		}

		private string CreateKey( int itemNumber, string propertyName )
		{
			return "SearchEngine" + itemNumber.ToString() + propertyName;
		}

		#endregion
	}
}
